#![warn(dead_code)]

use flax::{Schedule, World};
use anyhow::Result;
use crate::{
    core::events::Events,
    module::{Module, ModulesStack},
};

#[allow(dead_code)]
pub struct App {
    name: String,
    modules: ModulesStack,
    world: World,
    schedule: Schedule,
    events: Events,
    rx: flume::Receiver<AppEvent>,
    running: bool,
    event_cleanup_time: std::time::Duration,
}

impl App {
    pub fn new() -> Self {
        let mut events = Events::new();

        let (tx, rx) = flume::unbounded();
        events.subscribe_custom(tx);

        Self {
            name: "ZTest".into(),
            modules: ModulesStack::new(),
            world: World::new(),
            schedule: Schedule::default(),
            events,
            rx,
            running: false,
            event_cleanup_time: std::time::Duration::from_secs(60),
        }
    }

    pub fn run(&mut self) -> Result<()> {
        self.running = true;

        // self.schedule.execute_par(&mut self.world).unwrap();

        let world = &mut self.world;
        let events = &mut self.events;
        let frame_time = std::time::Duration::from_millis(16);

        for module in self.modules.iter_mut() {
            module.on_update(world, events, frame_time)?;
        }

        self.handle_events();

        Ok(())
    }

    pub fn handle_events(&mut self) {
        for event in self.rx.try_iter() {
            match event {
                AppEvent::Exit => self.running = false,
            }
        }
    }

    pub fn set_schedule(&mut self, schedule: Schedule) {
        self.schedule = schedule;
    }

    pub fn world(&self) -> &World {
        &self.world
    }

    pub fn world_mut(&mut self) -> &mut World {
        &mut self.world
    }

    pub fn events(&self) -> &Events {
        &self.events
    }

    pub fn events_mut(&mut self) -> &mut Events {
        &mut self.events
    }

    /// Pushes a layer from the provided init closure to to the top of the layer stack. The provided
    /// closure to construct the layer takes in the world and events.
    pub fn push_module<F, T>(&mut self, func: F)
    where
        F: FnOnce(&mut World, &mut Events) -> T,
        T: 'static + Module,
    {
        let module = func(&mut self.world, &mut self.events);
        self.modules.push(module);
    }

    /// Pushes a module from the provided init closure to to the top of the module stack. The provided
    /// closure to construct the module takes in the world and events, and may return an error which
    /// is propagated to the callee.
    pub fn try_push_module<F, T, E>(&mut self, func: F) -> Result<(), E>
    where
        F: FnOnce(&mut World, &mut Events) -> Result<T, E>,
        T: 'static + Module,
    {
        let module = func(&mut self.world, &mut self.events)?;
        self.modules.push(module);
        Ok(())
    }

    /// Inserts a module from the provided init closure to to the top of the module stack. The provided
    /// closure to construct the module takes in the world and events.
    pub fn insert_module<F, T>(&mut self, index: usize, func: F)
    where
        F: FnOnce(&mut World, &mut Events) -> T,
        T: 'static + Module,
    {
        let module = func(&mut self.world, &mut self.events);
        self.modules.insert(index, module);
    }

    /// Pushes a module from the provided init closure to to the top of the module stack. The provided
    /// closure to construct the module takes in the world and events, and may return an error which
    /// is propagated to the callee.
    pub fn try_insert_module<F, T, E>(&mut self, index: usize, func: F) -> Result<(), E>
    where
        F: FnOnce(&mut World, &mut Events) -> Result<T, E>,
        T: 'static + Module,
    {
        let module = func(&mut self.world, &mut self.events)?;
        self.modules.insert(index, module);
        Ok(())
    }
}

#[derive(Debug, Clone, Copy, PartialEq)]
#[allow(dead_code)]
pub enum AppEvent {
    Exit,
}

impl Default for App {
    fn default() -> Self {
        Self::new()
    }
}
