use flax::{Schedule, World};
use notify::{Config as NotifyConfig, INotifyWatcher, RecommendedWatcher, RecursiveMode, Watcher};
use serde::{Deserialize, Serialize};
use std::env::current_dir;

use crate::module::Module;

use self::{components::{notify_file_event, resources}, systems::{read_config_system, read_notify_events_system}};

pub mod components;
pub mod systems;

#[derive(Serialize, Deserialize, Debug, PartialEq, Eq)]
pub struct Config {
    pub asset_path: String,
}


#[allow(dead_code)]
pub struct ConfigModule {
    schedule: Schedule,
    watcher: INotifyWatcher,
    watcher_rx: std::sync::mpsc::Receiver<Result<notify::Event, notify::Error>>,
}

impl ConfigModule {
    pub fn new(_world: &mut World, _events: &mut crate::core::events::Events) -> Self {
        let (tx, rx) = std::sync::mpsc::channel();
        let mut watcher = RecommendedWatcher::new(tx, NotifyConfig::default().with_poll_interval(std::time::Duration::from_secs(2))).unwrap();

        watcher
            .watch(&current_dir().unwrap(), RecursiveMode::NonRecursive)
            .unwrap();

        let schedule = Schedule::builder()
            .with_system(read_config_system())
            .with_system(read_notify_events_system())
            .build();

        Self {
            schedule,
            watcher,
            watcher_rx: rx,
        }
    }
}

impl Module for ConfigModule {
    fn on_update(
        &mut self,
        world: &mut World,
        _events: &mut crate::core::events::Events,
        _frame_time: std::time::Duration,
    ) -> anyhow::Result<()> {
        self.schedule.execute_par(world).unwrap();

        if let Ok(event) = self.watcher_rx.recv() {
            match event {
                Ok(e) => {
                    world.set(resources(), notify_file_event(), e.clone()).unwrap();
                }
                Err(e) => println!("Watcher error. {}", e),
            }
        }

        Ok(())
    }
}
